Attribute VB_Name = "renderWireframe"

Option Explicit

Public Declare Function SetPixel Lib "gdi32" (ByVal hdc As Long, _
    ByVal x As Long, ByVal y As Long, ByVal crColor As Long) As Long

Public ourZBuffer As zBuffer

Public Type coord3d
  x As Double
  y As Double
  z As Double
End Type

Public Type wire
  begin As Long
  end As Long
  color As Long
End Type

Public Type mesh3d
  wires() As wire
  wiresCount As Long
  vertices() As coord3d
  verticesCount As Long
End Type

Public Type zLayer
  color As Long
  z As Double
End Type

Public Type zPixel
  layerCount As Long
  layers() As zLayer
End Type

Public Type zBuffer
  width As Long
  height As Long
  centerX As Long
  centerY As Long
  pixels() As zPixel
End Type

Function perspective(mesh As mesh3d, cameraZPos As Double) As mesh3d
  Dim i As Long, z As Double
  perspective.verticesCount = mesh.verticesCount
  ReDim perspective.vertices(1 To perspective.verticesCount) As coord3d
  perspective.wiresCount = mesh.wiresCount
  If perspective.wiresCount <> 0 Then
    ReDim perspective.wires(1 To perspective.wiresCount) As wire
  Else
    ReDim perspective.wires(1 To 1) As wire
  End If
  For i = 1 To perspective.wiresCount
    perspective.wires(i) = mesh.wires(i)
  Next i
  For i = 1 To perspective.verticesCount
    z = mesh.vertices(i).z
    perspective.vertices(i).x = (mesh.vertices(i).x * (z + cameraZPos)) / cameraZPos
    perspective.vertices(i).y = (mesh.vertices(i).y * (z + cameraZPos)) / cameraZPos
    perspective.vertices(i).z = z
  Next i
End Function

Function findExtVert(xQuest As Boolean, maxQuest As Boolean, _
    mesh As mesh3d) As Double
  Dim i As Long, maxVal As Double, minVal As Double
  If xQuest Then
    maxVal = minVal = mesh.vertices(1).x
  Else
    maxVal = minVal = mesh.vertices(1).y
  End If
  For i = 2 To mesh.verticesCount
    If xQuest Then
      If maxQuest Then
        If mesh.vertices(i).x > maxVal Then maxVal = mesh.vertices(i).x
      Else
        If mesh.vertices(i).x < minVal Then minVal = mesh.vertices(i).x
      End If
    Else
      If maxQuest Then
        If mesh.vertices(i).y > maxVal Then maxVal = mesh.vertices(i).y
      Else
        If mesh.vertices(i).y < minVal Then minVal = mesh.vertices(i).y
      End If
    End If
  Next i
  If maxQuest Then
    findExtVert = maxVal
  Else
    findExtVert = minVal
  End If
End Function

Sub initPixel(x As Long, y As Long, z As Double, color As Long)
  Dim i As Long
  For i = 1 To ourZBuffer.pixels(x, y).layerCount
    If ourZBuffer.pixels(x, y).layers(i).z = z Then Exit Sub
  Next i
  ourZBuffer.pixels(x, y).layerCount = ourZBuffer.pixels(x, y).layerCount + 1
  ReDim Preserve ourZBuffer.pixels(x, y).layers(1 To _
      ourZBuffer.pixels(x, y).layerCount) As zLayer
  ourZBuffer.pixels(x, y).layers(ourZBuffer.pixels(x, y).layerCount).color = color
  ourZBuffer.pixels(x, y).layers(ourZBuffer.pixels(x, y).layerCount).z = z
End Sub

Sub drawWire(ByVal X1 As Long, ByVal Y1 As Long, Z1 As Double, _
    ByVal X2 As Long, ByVal Y2 As Long, Z2 As Double, color As Long)
  Dim i As Long, xp As Long, yp As Long, zp As Double
  If Not (X2 - X1 = 0 And Y2 - Y1 = 0) Then
    If Abs(X2 - X1) >= Abs(Y2 - Y1) Then
      For i = X1 To X2 Step Sgn(X2 - X1)
        xp = i
        yp = Round(Y1 + ((Y2 - Y1) * Abs((i - X1) / (X2 - X1))))
        zp = Z1 + ((Z2 - Z1) * Abs((i - X1) / (X2 - X1)))
        initPixel xp, yp, zp, color
      Next i
    Else
      For i = Y1 To Y2 Step Sgn(Y2 - Y1)
        xp = Round(X1 + ((X2 - X1) * Abs((i - Y1) / (Y2 - Y1))))
        yp = i
        zp = Z1 + ((Z2 - Z1) * Abs((i - Y1) / (Y2 - Y1)))
        initPixel xp, yp, zp - 0.01, color
      Next i
    End If
  Else
    initPixel X1, Y1, CInt(Z1 < Z2) * Z1 + CInt(Z2 < Z1) * Z2, color
  End If
End Sub

Sub renderMesh3d(hdc As Long, mesh As mesh3d, _
    centerX As Long, centerY As Long, scaleFactor As Double, _
    cameraZPos As Double, perspectiveTransformation As Boolean)
  Dim i As Long, j As Long, k As Long, _
      maxX As Double, maxY As Double, minX As Double, minY As Double, _
      zMin As Double, minNum As Long
  If perspectiveTransformation Then mesh = perspective(mesh, cameraZPos)
  maxX = findExtVert(True, True, mesh) * scaleFactor
  maxY = findExtVert(False, True, mesh) * scaleFactor
  minX = findExtVert(True, False, mesh) * scaleFactor
  minY = findExtVert(False, False, mesh) * scaleFactor
  ourZBuffer.width = maxX - minX
  ourZBuffer.height = maxY - minY
  ourZBuffer.centerX = -minX
  ourZBuffer.centerY = maxY
  ReDim ourZBuffer.pixels(0 To ourZBuffer.width, 0 To ourZBuffer.height) As zPixel
  For i = 1 To mesh.wiresCount
    drawWire (mesh.vertices(mesh.wires(i).begin).x * scaleFactor - minX), _
        (maxY - mesh.vertices(mesh.wires(i).begin).y * scaleFactor), _
        mesh.vertices(mesh.wires(i).begin).z, _
        (mesh.vertices(mesh.wires(i).end).x * scaleFactor - minX), _
        (maxY - mesh.vertices(mesh.wires(i).end).y * scaleFactor), _
        mesh.vertices(mesh.wires(i).end).z, _
        mesh.wires(i).color
  Next i
  For j = 0 To ourZBuffer.height
    For i = 0 To ourZBuffer.width
      If ourZBuffer.pixels(i, j).layerCount > 0 Then
        zMin = ourZBuffer.pixels(i, j).layers(1).z
        minNum = 1
        For k = 2 To ourZBuffer.pixels(i, j).layerCount
          If ourZBuffer.pixels(i, j).layers(k).z < zMin _
              And ourZBuffer.pixels(i, j).layers(k).z > cameraZPos Then
            zMin = ourZBuffer.pixels(i, j).layers(k).z
            minNum = k
          End If
        Next k
        SetPixel hdc, centerX - ourZBuffer.centerX + i, _
            centerY - ourZBuffer.centerY + j, _
            ourZBuffer.pixels(i, j).layers(minNum).color
      End If
    Next i
  Next j
End Sub
